from bitcoin_tools.FieldElement import FieldElement
from bitcoin_tools.Point import Point
from bitcoin_tools.helper import sha256_ripemd160, double_sha256, encode_base58
from unittest import TestCase
from binascii import hexlify, unhexlify
import math

P = 2 ** 256 - 2 ** 32 - 977
N = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141
A = 0
B = 7


class S256Field(FieldElement):
    def __init__(self, num, prime=None):
        super().__init__(num=num, prime=P)

    def hex(self):
        return '{:x}'.format(self.num).zfill(64)

    def __repr__(self):
        return self.hex()

    def sqrt(self):
        return self ** ((P + 1) // 4)


class S256Point(Point):
    bits = 256

    def __init__(self, x, y, a=None, b=None):
        a, b = S256Field(A), S256Field(B)

        if x is None:
            super().__init__(x=None, y=None, a=a, b=b)
        elif type(x) == int:
            super().__init__(S256Field(x), S256Field(y), a=a, b=b)
        else:
            super().__init__(x=x, y=y, a=a, b=b)

    def __repr__(self):
        if self.x is None:
            return 'Point(infinity)'
        else:
            return 'Point({},{})'.format(self.x, self.y)

    def __rmul__(self, scalar):
        # Binary Expansion on current
        current = self

        result = S256Point(None, None)

        # Iterating in range of 256 and doubling each time ->
        # when there is a 1 at the end of the binary number ->
        # add to result
        for _ in range(self.bits):
            # If last bit is 1, add current to self (adding -> p1 + p2 = p3)
            if scalar & 1:
                # Calling super class (Point) to calculate the next point on the curve  ->
                # adding result(x,y) + current (x,y)
                result += current

            # Calling super class (Point) to point double current(x,y) * current(x,y)
            current += current

            # Shifting bits of scalar to the right by 1 and assigning new value
            # Decrementing scalar by /2
            scalar >>= 1

        return result

    def get_sec(self, compressed=True):

        if self.x is None or self.y is None:
            raise RuntimeError("x is None")

        x = self.x.num.to_bytes(32, 'big')
        y = self.y.num.to_bytes(32, 'big')

        if compressed:
            if self.y.num % 2 == 0:
                prefix = b'\x02'
            else:
                prefix = b'\x03'

            sec = prefix + x

        else:
            prefix = b'\x04'

            sec = prefix + x + y

        return hexlify(sec)

    def get_address(self, sec, mainnet=False):

        if type(sec) is not bytes:
            raise RuntimeError("sec must be passed as bytes format")

        # Need to turn sec into bytes before hashing
        sec_bytes = unhexlify(sec)

        if mainnet:
            prefix = b'\x00'
        else:
            prefix = b'\x6f'


        # SHA256 -> RIMPEMD160
        hashed_sec_bytes = sha256_ripemd160(sec_bytes)

        # Concatenate prefix and hashed sec bytes
        raw = prefix + hashed_sec_bytes

        # Double SHA256 raw and take the first 4 bytes as checksum
        checksum = double_sha256(raw)[:4]

        # Encode Base58 raw and checksum to generate accepted address
        address = encode_base58(raw + checksum)

        return address.decode('ascii')

    @classmethod
    def parse(cls, sec_bin):
        '''returns a Point object from a compressed sec binary (not hex)
        '''
        # Uncompressed SEC
        if sec_bin[0] == 4:
            x = int(hexlify(sec_bin[1:33]), 16)
            y = int(hexlify(sec_bin[33:65]), 16)
            return S256Point(x=x, y=y)

        # Comprssed SEC
        is_even = sec_bin[0] == 2
        x = S256Field(int(hexlify(sec_bin[1:]), 16))

        # right side of the equation y^2 = x^3 + 7
        alpha = x ** 3 + S256Field(B)
        # solve for left side
        beta = alpha.sqrt()

        if beta.num % 2 == 0:
            even_beta = beta
            odd_beta = S256Field(P - beta.num)
        else:
            even_beta = S256Field(P - beta.num)
            odd_beta = beta
        if is_even:
            return S256Point(x, even_beta)
        else:
            return S256Point(x, odd_beta)

    def verify(self, z, sig):
        # remember sig.r and sig.s are the main things we're checking
        # remember 1/s = pow(s, N-2, N)
        s_inv = pow(sig.s, N - 2, N)
        # u = z / s
        u = z * s_inv % N
        # v = r / s
        v = sig.r * s_inv % N
        # u*G + v*P should have as the x coordinate, r
        total = u * G + v * self
        return total.x.num == sig.r


G = S256Point(
    0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798,
    0x483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8)




import json
import requests

# 1st case -> move to its own Class?
# 5757068b376143b8aa5f7fc137dc2351
block_cypher_url = "https://api.blockcypher.com/v1/btc/test3"
block_cypher_token = "?token=c355cd4305b042a980c6b83c4e21bb4d"

# 2nd case -> move to its own Class?
block_trail_url = "https://api.blocktrail.com/v1/tBTC"
block_trail_token = "?api_key=8a2a363be9edf703cba639d7e3f8a1831978f483"

def ping():
    return requests.get(block_cypher_url)


def get_balance(address):
    balance_url = "/addrs/{}/balance".format(address)
    response = requests.get(block_cypher_url + balance_url)

    if response.status_code != 200:
        raise RuntimeError("The server returned an error: {}".format(response.json()))

    return response


def send_tx(raw_tx):
    send_tx_url = "/txs/push"

    json_tx = create_json_tx(raw_tx)

    response = requests.post(block_cypher_url + send_tx_url + block_cypher_token, data=json_tx)

    if response.status_code == 400:
        raise RuntimeError("Error sending the tx: {}".format(response.json()))

    return response


def get_transaction(transaction_hash):
    get_tx_url = "/txs/{}".format(transaction_hash)

    response = requests.get(block_cypher_url + get_tx_url)

    return response


def decode_transaction(raw_tx):
    decode_tx_url = "/txs/decode"

    json_tx = create_json_tx(raw_tx)

    response = requests.post(block_cypher_url + decode_tx_url + block_cypher_token, data=json_tx)

    return response


def create_json_tx(raw_tx):
    tx_object = {"tx": raw_tx}
    return json.dumps(tx_object)

def request_balance(address):
    # 1st case
    request_balance_url = "/addrs/{}/balance".format(address)

    try:
        response = requests.get(block_cypher_url + request_balance_url)
        # if response.status_code != 200:
            # raise RuntimeError("The server returned an error: {}".format(response.json()))
    except:
        raise RuntimeError("API Limit reached: {}".format(response.json()))

    # 2nd case
    # Block trail does not have a balance request
    try:
        request_UTXOs_url = "/address/{}/unspent-outputs".format(address)

        response = requests.get(block_trail_url + request_UTXOs_url + block_trail_token)
        # if response.status_code != 200:
            # raise RuntimeError("The server returned an error: {}".format(response.json()))
    except:
        raise RuntimeError("API Limit reached: {}".format(response.json()))
        
    print("Response from Block Trail: {}".format(response))

    return response

def request_UTXOs(address):

    response = None
    schema = None

    # 1st case
    # try: 
    #     request_UTXOs_url = "/addrs/{}".format(address)

    #     response = requests.get(block_cypher_url + request_UTXOs_url)

    #     # if response.status_code != 200:
    #         # raise RuntimeError("The server returned an error: {}".format(response.json()))

    #     schema = 'block_cypher'
    # except:
    #     raise RuntimeError("API Limit reached: {}".format(response.json()))


    # 2nd case
    try: 
        request_UTXOs_url = "/address/{}/unspent-outputs".format(address)
        response = requests.get(block_trail_url + request_UTXOs_url + block_trail_token)
        print("Block_trail resp: {}".format(response.json()))

        # if response.status_code != 200:
            # raise RuntimeError("The server returned an error: {}".format(response.json()))

        schema = 'block_trail'

    except:
        raise RuntimeError("API Limit reached: {}".format(response.json()))    

    return (response, schema)


import secrets
from bitcoin_tools.Signature import Signature
from bitcoin_tools.S256Point import G, N, P, A, B

class ECC:
    def generate_priv_key(self):
        return secrets.randbelow(N)

    def generate_pub_key(self, priv_key):
        return priv_key * G

    def is_on_curve(self, x, y):
        return (y ** 2) % P == (x ** 3 + A + B) % P

    # TODO: Examine and review
    # TODO: K needs to have higher entropy, use tx as seed
    def generate_signature(self, priv_key):
        z = secrets.randbelow(2 ** 256)
        k = secrets.randbelow(2 ** 256)
        r = (k * G).x.num

        sig = (z + r * priv_key) * pow(k, N - 2, N) % N

        return Signature(z, r, sig)






from bitcoin_tools.S256Point import *
from bitcoin_tools.FieldElement import *
from bitcoin_tools.Point import *
from bitcoin_tools.blockchain_explorer_helper import *
from bitcoin_tools.ECC import *
from bitcoin_tools.helper import *
from bitcoin_tools.PrivateKey import *
from bitcoin_tools.Script import *
from bitcoin_tools.Tx import *
from bitcoin_tools.UTXO import *
from bitcoin_tools.Signature import *


# This class is the basis for performing modulo arithmetic operations on x,y points on the curve

class FieldElement:

    def __init__(self, num, prime):
        self.num = num
        self.prime = prime

        if self.num >= self.prime or self.num < 0:
            raise RuntimeError("Num is not within the field of: {0}".format(self.prime - 1))

    # Checks for equality between self and other
    def __eq__(self, other):
        # if other is None:
        #     return True

        return self.num == other.num and self.prime == other.prime

    def __ne__(self, other):
        if other is None:
            return True
        return self.num != other.num or self.prime != other.prime

    # The operations performed when self and other are added together
    def __add__(self, other):
        self.check_primes_the_same(other.prime)

        num = (self.num + other.num) % self.prime

        # Returns an instance of the FieldElement class
        return self.__class__(num, self.prime)

    def __sub__(self, other):
        self.check_primes_the_same(other.prime)

        num = (self.num - other.num) % self.prime

        return self.__class__(num, self.prime)

    def __mul__(self, other):
        self.check_primes_the_same(other.prime)

        num = (self.num * other.num) % self.prime

        return self.__class__(num, self.prime)

    def __pow__(self, power):
        num = pow(self.num, power % (self.prime - 1), self.prime)
        return self.__class__(num, self.prime)

    def __truediv__(self, other):
        self.check_primes_the_same(other.prime)
        num = (self.num * pow(other.num, self.prime - 2, self.prime)) % self.prime

        return self.__class__(num, self.prime)

    def __rmul__(self, scalar):
        num = (scalar * self.num) % self.prime
        return self.__class__(num, self.prime)

    # Function to check if primes are the same number
    def check_primes_the_same(self, other_prime):
        if self.prime != other_prime:
            raise RuntimeError("Primes are different")



from binascii import unhexlify, hexlify
from io import BytesIO
from bitcoin_tools.PrivateKey import PrivateKey
from bitcoin_tools.helper import decode_base58, p2pkh_script, SIGHASH_ALL

OP_CODES = {
    0: 'OP_0',
    76: 'OP_PUSHDATA1',
    77: 'OP_PUSHDATA2',
    78: 'OP_PUSHDATA4',
    79: 'OP_1NEGATE',
    80: 'OP_RESERVED',
    81: 'OP_1',
    82: 'OP_2',
    83: 'OP_3',
    84: 'OP_4',
    85: 'OP_5',
    86: 'OP_6',
    87: 'OP_7',
    88: 'OP_8',
    89: 'OP_9',
    90: 'OP_10',
    91: 'OP_11',
    92: 'OP_12',
    93: 'OP_13',
    94: 'OP_14',
    95: 'OP_15',
    96: 'OP_16',
    97: 'OP_NOP',
    98: 'OP_VER',
    99: 'OP_IF',
    100: 'OP_NOTIF',
    101: 'OP_VERIF',
    102: 'OP_VERNOTIF',
    103: 'OP_ELSE',
    104: 'OP_ENDIF',
    105: 'OP_VERIFY',
    106: 'OP_RETURN',
    107: 'OP_TOALTSTACK',
    108: 'OP_FROMALTSTACK',
    109: 'OP_2DROP',
    110: 'OP_2DUP',
    111: 'OP_3DUP',
    112: 'OP_2OVER',
    113: 'OP_2ROT',
    114: 'OP_2SWAP',
    115: 'OP_IFDUP',
    116: 'OP_DEPTH',
    117: 'OP_DROP',
    118: 'OP_DUP',
    119: 'OP_NIP',
    120: 'OP_OVER',
    121: 'OP_PICK',
    122: 'OP_ROLL',
    123: 'OP_ROT',
    124: 'OP_SWAP',
    125: 'OP_TUCK',
    126: 'OP_CAT',
    127: 'OP_SUBSTR',
    128: 'OP_LEFT',
    129: 'OP_RIGHT',
    130: 'OP_SIZE',
    131: 'OP_INVERT',
    132: 'OP_AND',
    133: 'OP_OR',
    134: 'OP_XOR',
    135: 'OP_EQUAL',
    136: 'OP_EQUALVERIFY',
    137: 'OP_RESERVED1',
    138: 'OP_RESERVED2',
    139: 'OP_1ADD',
    140: 'OP_1SUB',
    141: 'OP_2MUL',
    142: 'OP_2DIV',
    143: 'OP_NEGATE',
    144: 'OP_ABS',
    145: 'OP_NOT',
    146: 'OP_0NOTEQUAL',
    147: 'OP_ADD',
    148: 'OP_SUB',
    149: 'OP_MUL',
    150: 'OP_DIV',
    151: 'OP_MOD',
    152: 'OP_LSHIFT',
    153: 'OP_RSHIFT',
    154: 'OP_BOOLAND',
    155: 'OP_BOOLOR',
    156: 'OP_NUMEQUAL',
    157: 'OP_NUMEQUALVERIFY',
    158: 'OP_NUMNOTEQUAL',
    159: 'OP_LESSTHAN',
    160: 'OP_GREATERTHAN',
    161: 'OP_LESSTHANOREQUAL',
    162: 'OP_GREATERTHANOREQUAL',
    163: 'OP_MIN',
    164: 'OP_MAX',
    165: 'OP_WITHIN',
    166: 'OP_RIPEMD160',
    167: 'OP_SHA1',
    168: 'OP_SHA256',
    169: 'OP_HASH160',
    170: 'OP_HASH256',
    171: 'OP_CODESEPARATOR',
    172: 'OP_CHECKSIG',
    173: 'OP_CHECKSIGVERIFY',
    174: 'OP_CHECKMULTISIG',
    175: 'OP_CHECKMULTISIGVERIFY',
    176: 'OP_NOP1',
    177: 'OP_NOP2',
    177: 'OP_CHECKLOCKTIMEVERIFY',
    178: 'OP_NOP3',
    178: 'OP_CHECKSEQUENCEVERIFY',
    179: 'OP_NOP4',
    180: 'OP_NOP5',
    181: 'OP_NOP6',
    182: 'OP_NOP7',
    183: 'OP_NOP8',
    184: 'OP_NOP9',
    185: 'OP_NOP10',
    252: 'OP_NULLDATA',
    253: 'OP_PUBKEYHASH',
    254: 'OP_PUBKEY',
    255: 'OP_INVALIDOPCODE',
}


class Script:
    def __init__(self, elements):
        self.elements = elements

    @classmethod
    def parse(cls, raw):
        stream = BytesIO(raw)

        elements = []

        current = stream.read(1)

        while current != b'':
            # Read the int representation of the byte
            op_code = current[0]

            if 0 < op_code <= 75:
                # Read the next set of bytes as the length of the op_code
                # which should be the length of the pub key hash and add it to the list
                elements.append(stream.read(op_code))
            else:
                elements.append(op_code)

            current = stream.read(1)
        return cls(elements)

    # Returns a print format for the object
    def __repr__(self):
        result = ''
        for element in self.elements:
            if type(element) == int:
                result += '{} '.format(OP_CODES[element])
            else:
                result += '{} '.format(hexlify(element))
        return result

    def type(self):
        if len(self.elements) == 0:
            return 'blank'
        elif self.elements[0] == 0x76 \
                and self.elements[1] == 0xa9 \
                and type(self.elements[2]) == bytes \
                and len(self.elements[2]) == 0x14 \
                and self.elements[3] == 0x88 \
                and self.elements[4] == 0xac:

            # hex/int
            # script_pubkey = <76/118 : OP_DUP> <a9/169 : OP_HASH160> <14/20 : Length of hash> <88/136 : OP_EQUAL_VERIFY> <ac/172 : OP_CHECKSIG>
            return 'p2pkh'
        elif self.elements[0] == 0xa9 \
                and type(self.elements[1]) is bytes \
                and len(self.elements[1]) == 0x14 \
                and self.elements[2] == 0x87:

            # <a9 : OP_HASH16-> <14 : Length of hash> < hash > <87 : OP_EQUAL>
            return 'p2sh'
        elif type(self.elements[0]) == bytes \
                and len(self.elements[0]) in (0x47, 0x48, 0x49) \
                and type(self.elements[1]) == bytes \
                and len(self.elements[1]) in (0x21, 0x41):
            # p2pkh scriptSig:
            # <signature> <pubkey>
            return 'p2pkh sig'
        elif len(self.elements) > 1 \
                and type(self.elements[1]) == bytes \
                and len(self.elements[1]) in (0x47, 0x48, 0x49) \
                and self.elements[-1][-1] == 0xae:
            # HACK: assumes p2sh is a multisig
            # p2sh multisig:
            # <x> <sig1> ... <sigm> <redeemscript ends with OP_CHECKMULTISIG>
            return 'p2sh sig'

    def serialize(self):
        result = b''
        for each_element in self.elements:
            # if element is an int, it's an OP_CODEgit
            if type(each_element) == int:
                result += bytes([each_element])
            else:
                # Element is the hash160 of the sec pub key, prefix it with the length of bytes and then the h160
                result += bytes([len(each_element)]) + each_element

        return result

    def der_signature(self, index=0):
        '''index isn't used for p2pkh, for p2sh, means one of m sigs'''
        sig_type = self.type()
        if sig_type == 'p2pkh sig':
            return self.elements[0]
        elif sig_type == 'p2sh sig':
            return self.elements[index + 1]
        else:
            raise RuntimeError('script type needs to be p2pkh sig or p2sh sig')

    def sec_pubkey(self, index=0):
        '''index isn't used for p2pkh, for p2sh, means one of n pubkeys'''
        sig_type = self.type()
        if sig_type == 'p2pkh sig':
            # Returns the Pub Key (A SEC public key)
            return self.elements[1]
        elif sig_type == 'p2sh sig':
            # HACK: assumes p2sh is a multisig
            redeem_script = Script.parse(self.elements[-1])
            return redeem_script.elements[index + 1]

    def redeem_script(self):
        sig_type = self.type()
        #If the type is p2sh signature, return the last object which is the redeem script
        if sig_type == 'p2sh sig':
            return self.elements[-1]
        else:
            raise RuntimeError('script types needs to be p2sh sig')






from bitcoin_tools.FieldElement import FieldElement

class Point:
    def __init__(self, x, y, a, b):
        self.x = x
        self.y = y
        self.a = a
        self.b = b

        if self.x is None and self.y is None:
            return

        if self.y ** 2 != self.x ** 3 + self.a * self.x + self.b:
            raise RuntimeError("Points not on the curve")

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y and \
               self.a == other.a and self.b == other.b

    def __ne__(self, other):
        return self.x != other.x or self.y != other.y or \
               self.a != other.a or self.b != other.b

    def __repr__(self):
        if self.x is None:
            return 'Point(infinity)'
        else:
            return 'Point({},{})'.format(self.x, self.y)

    def __add__(self, other):
        if self.a != other.a or self.b != other.b:
            raise RuntimeError("Points are using different curves")

        if self.x is None:
            return other

        if other.x is None:
            return self

        # Check for reaching point at infinity
        # X are the same point but y is not, this would be a perfectly horizontal line on the curve
        if self.x == other.x and self.y != other.y:
            return self.__class__(None, None, self.a, self.b)

        # Calculate if self.x != other.x
        if self.x != other.x:
            slope = (self.y - other.y) / (self.x - other.x)

            # x3 = s**2 -x1 -x2
            x3 = slope ** 2 - self.x - other.x

            # y3 = slope * (x1 - x3) - y1
            y3 = slope * (self.x - x3) - self.y

            return self.__class__(x3, y3, self.a, self.b)
        else:
            # Point Doubling because x is the same value
            # Slope = (3* x1 ** 2 +a) / (2 * y1)
            slope = (3 * self.x ** 2 + self.a) / (2 * self.y)

            # x3 = slope**2 - 2*x1
            x3 = slope ** 2 - 2 * self.x

            # y3 = slope * (x1 - x3) - y1
            y3 = slope * (self.x - x3) - self.y

            return self.__class__(x3, y3, self.a, self.b)

    def __rmul__(self, scalar):
        product = self.__class__(None, None, self.a, self.b)

        for _ in range(scalar):
            product += self

        return product



import hashlib
from binascii import hexlify, unhexlify
from unittest import TestCase

SIGHASH_ALL = 1
SIGHASH_NONE = 2
SIGHASH_SINGLE = 3
BASE58_ALPHABET = b'123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


def encode_base58(s):
    # Find the amount of leading 0 bytes
    count = 0
    for c in s:
        if c == 0:
            count += 1
        else:
            break

    # There are no 0's in base58,
    # Replace the leading 0's with 1's,
    # 1's represent 0's in base58
    prefix = b'1' * count

    # Convert hexadecimal value of s to base-16 integer
    num = int(hexlify(s), 16)

    # Mutable byte array from 0 to 256
    result = bytearray()

    # While the base 16 integer is greater than 0
    while num > 0:
        # divmod returns the quotient and remainder
        # Divide num by 58
        num, mod = divmod(num, 58)

        # Using mod as an index, retrieve corresponding letter/number
        # Insert this value at index 0 in result bytearray
        # Inserting at 0, doesn't override but shifts the the rest of the array by 1
        result.insert(0, BASE58_ALPHABET[mod])

    return prefix + bytes(result)


def encode_base58_checksum(s):
    return encode_base58(s + double_sha256(s)[:4]).decode('ascii')


def double_sha256(s):
    return hashlib.sha256(hashlib.sha256(s).digest()).digest()


def sha256_ripemd160(s):
    return hashlib.new('ripemd160', hashlib.sha256(s).digest()).digest()


def little_endian_to_int(b):
    return int.from_bytes(b, 'little')


def int_to_little_endian(num, length):
    return int.to_bytes(num, length, 'little')


# TODO: NEED to understand varint types better
def encode_varint(i):
    # Encodes an integer to varint
    # if i is less than 253 (int)
    if i < 0xfd:
        return bytes([i])
    elif i < 0x10000:
        return b'\xfd' + int_to_little_endian(i, 2)
    elif i < 0x100000000:
        return b'\xfe' + int_to_little_endian(i, 4)
    elif i < 0x10000000000000000:
        return b'\xff' + int_to_little_endian(i, 8)
    else:
        raise RuntimeError('integer too large: {}'.format(i))


def read_varint(s):
    i = s.read(1)[0]

    if i == 0xfd:
        # 0xfd means the next two bytes are the number
        return little_endian_to_int(s.read(2))
    elif i == 0xfe:
        # 0xfe means that the next four bytes are the number
        return little_endian_to_int(s.read(4))
    elif i == 0xff:
        # 0xff means the next eight bytes are the number
        return little_endian_to_int(s.read(8))
    else:
        # i is just the number
        return i


def satoshi_to_bitcoin(satoshi):
    return float(satoshi / 100000000)


def bitcoin_to_satoshi(bitcoin):
    return int(bitcoin * 100000000)


def decode_base58(address):
    num = 0
    # Iterate over each string, in the address, c is the integer value of the ascii character
    for c in address.encode('ascii'):
        num *= 58
        num += BASE58_ALPHABET.index(c)

    combined = num.to_bytes(25, byteorder='big')
    checksum = combined[-4:]

    # Check that combined bytes, double_sha256() -> take the last 4 bytes of combined and then take the first 4 bytes
    # This should equal checksum
    if double_sha256(combined[:-4])[:4] != checksum:
        raise RuntimeError('malformed address: {} {}'.format(checksum, double_sha256(combined)[:4]))

    return combined[1:-4]


def p2pkh_script(h160):
    # Takes the h160 of an address and inserts it into the p2pkh scriptPubKey format
    return b'\x76\xa9\x14' + h160 + b'\x88\xac'

def p2sh_script(h160):
    # Takes the h160 of an address and inserts it into p2sh scriptPubKey format
    return b'\xa9\x14' + h160 + b'\x87'

def generate_reedemScript(owners_sec, list_of_pub_keys):
    if len(list_of_pub_keys) < 1:
        raise RuntimeError("No public keys passed")

    unhexed_sec = unhexlify(owners_sec)
    len_of_unhexed_sec = bytes([len(unhexed_sec)])

    result = b'\x52' + len_of_unhexed_sec + unhexed_sec

    for pub_key in list_of_pub_keys:
        result += bytes([len(pub_key)])
        result += pub_key

    result += b'\x52' + b'\xae'

    return result

def generate_p2sh_address(redeem_script, mainnet=False):
        if mainnet:
            prefix = b'\x05'
        else:
            prefix = b'\xc4'

        h160 = sha256_ripemd160(redeem_script)
        address = encode_base58_checksum(prefix + h160)

        return address

def generate_p2pkh_pub_key(address):
    h160 = decode_base58(address)

    return p2pkh_script(h160)

def generate_p2sh_pub_key(address):
    h160 = decode_base58(address)

    return p2sh_script(h160)




#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(name='bitcoin_tools',
      version='0.0.12',
      description='Bitcoin Tools Library, WIP: CURRENTLY ONLY USED FOR TESTNET',
      author='Christopher Coverdale',
      author_email='chris.coverdale24@gmail.com',
      url='https://github.com/ccdle12/python-bitcoin-tools',
      include_package_data=True,
      packages=['bitcoin_tools'],
      )



from bitcoin_tools.S256Point import N, G, B, P, A
from bitcoin_tools.helper import encode_base58_checksum
from bitcoin_tools.Signature import Signature
from random import randint
import secrets

class PrivateKey:
    def __init__(self, secret=None):

        if secret is None:
            self.secret = secrets.randbelow(N)
        else:
            self.secret = secret

        self.public_key = self.secret * G

        if self.is_on_curve() is False:
            raise RuntimeError("Generated Public Point is not on Curve, do not use this private key")

    def is_on_curve(self):
        return (self.public_key.y.num ** 2) % P == (self.public_key.x.num ** 3 + A + B) % P

    def get_WIF(self, compressed=True, mainnet=True):

        if mainnet:
            prefix = b'\x80'
        else:
            prefix = b'\xef'

        s = self.secret.to_bytes(32, 'big')

        wif = prefix + s

        if compressed:
            wif = wif + b'\x01'

        return encode_base58_checksum(wif)

    @classmethod
    def import_private_key(cls, secret):
        return cls(secret)

    def sign(self, z):

        k = randint(0,  2**256)
        r = (k * G).x.num

        s = (z + r * self.secret) * pow(k, N - 2, N) % N

        if s > N // 2:
            s = N - s

        return Signature(r, s)



from bitcoin_tools.helper import little_endian_to_int, read_varint, satoshi_to_bitcoin, int_to_little_endian, encode_varint, double_sha256, SIGHASH_ALL, decode_base58
from bitcoin_tools.Signature import Signature
from bitcoin_tools.Script import Script
from bitcoin_tools.S256Point import S256Point
import requests
from io import BytesIO
from binascii import unhexlify, hexlify


class Tx:
    def __init__(self, version, tx_ins, tx_outs, locktime, testnet=True):
        self.version = version
        self.tx_ins = tx_ins
        self.tx_outs = tx_outs
        self.locktime = locktime
        self.testnet = testnet

    def __repr__(self):
        tx_ins = ''
        for tx_in in self.tx_ins:
            tx_ins += tx_in.__repr__() + '\n'

        tx_outs = ''
        for tx_out in self.tx_outs:
            tx_outs += tx_out.__repr__() + '\n'

        return 'version: {}\ntx_ins:\n{}\ntx_outs:\n{}\nlocktime: {}\n'.format(
            self.version,
            tx_ins,
            tx_outs,
            self.locktime,
        )

    # Allows method to be called without instantiating class, cls points to the class not the object instance
    # MUST pass in RAW BYTES
    @classmethod
    def parse(cls, transaction_bytes):
        if type(transaction_bytes) is str:
            transaction_bytes = unhexlify(transaction_bytes)

        # Create a stream of bytes
        stream = BytesIO(transaction_bytes)

        # Retrieve the version as int
        # Read the first 4 bytes of the stream, interpret as int, little endian architecture
        version = little_endian_to_int(stream.read(4))

        # Number of inputs is a varint
        num_inputs = read_varint(stream)

        inputs = []
        for _ in range(num_inputs):
            inputs.append(TxIn.parse(stream))

        # Reads number of outputs as a varint to int
        num_outputs = read_varint(stream)
        outputs = []
        for _ in range(num_outputs):
            outputs.append(TxOut.parse(stream))

        # Read the locktime as int, little endian
        locktime = little_endian_to_int(stream.read(4))

        return cls(version, inputs, outputs, locktime)

    # TODO: Serializes data in object and returns as bytes
    def serialize(self):
        result = int_to_little_endian(self.version, 4)

        result += encode_varint(len(self.tx_ins))

        for tx_in in self.tx_ins:
            result += tx_in.serialize()

        result += encode_varint(len(self.tx_outs))

        for tx_out in self.tx_outs:
            result += tx_out.serialize()

        result += int_to_little_endian(self.locktime, 4)

        return result

    def calculate_fee(self):
        input_amount = 0
        output_amount = 0

        for tx_in in self.tx_ins:
            input_amount += tx_in.value()

        for tx_out in self.tx_outs:
            output_amount += tx_out.amount

        return input_amount - output_amount

    def sig_hash(self, input_index, hash_type):
        '''Returns the integer representation of the hash that needs to get
        signed for index input_index'''
        # create a new set of tx_ins (alt_tx_ins)
        alt_tx_ins = []

        # iterate over self.tx_ins
        for tx_in in self.tx_ins:
            # create a new TxIn that has a blank script_sig (b'') and add to alt_tx_ins
            print("Prev Transaction of input: {}".format(hexlify(tx_in.prev_hash)))
            alt_tx_ins.append(TxIn(
                prev_hash=tx_in.prev_hash,
                prev_index=tx_in.prev_index,
                script_sig=b'',
                sequence=tx_in.sequence,
            ))

        # grab the input at the input_index
        # "1b562bc7c059af8a61bec721bae5c8f47d929389049b38525045e6330ac7acf2:1"
        signing_input = alt_tx_ins[input_index]
        print("SIGNING INPUT: {}".format(signing_input))

        # grab the script_pubkey of the input
        script_pubkey = signing_input.script_pubkey(self.testnet)
        print("Script pubkey: {}".format(script_pubkey))

        # Check the sig type
        sig_type = script_pubkey.type()

        if sig_type == 'p2pkh':
            signing_input.script_sig = script_pubkey
        elif sig_type == 'p2sh':
            # Get the tx input at index passed of the tx_object
            current_input = self.tx_ins[input_index]
            # current_input.script_sig = Script.parse(redeem_script)
            # print("REDEEM SCRIPT Sig Passed: {}".format(current_input.redeem_script()))

            # signing_input.script_sig = Script.parse(current_input.redeem_script())
        else:
            raise RuntimeError('no valid sig_type')

        # create an alternate transaction with the modified tx_ins
        alt_tx = self.__class__(
            version=self.version,
            tx_ins=alt_tx_ins,
            tx_outs=self.tx_outs,
            locktime=self.locktime)

        # add the hash_type int 4 bytes, little endian
        result = alt_tx.serialize() + int_to_little_endian(hash_type, 4)

        # get the double_sha256 of the tx serialization
        s256 = double_sha256(result)

        # convert this to a big-endian integer using int.from_bytes(x, 'big')
        return int.from_bytes(s256, 'big')

    def validate_signature(self, index_pos):
        # Parse the transaction

        # Get the input at index_pos provided
        tx_in = self.tx_ins[index_pos]

        # Get the der signature and hash type
        der, hash_type = tx_in.der_signature()

        # Parse the der signature
        signature = Signature.parse(der)

        # Get the sec pub key of the input
        sec = tx_in.sec_pubkey()

        # Parse the sec_pubkey to return x,y points
        point = S256Point.parse(sec)

        # Use sig_has method on transaction to turn transaction into z
        sig_hash = self.sig_hash(index_pos, hash_type)

        # Verify both x co-ordinates are the same in signed hash and signature
        return point.verify(sig_hash, signature)


class TxIn:
    cache = {}

    def __init__(self, prev_hash, prev_index, script_sig, sequence):
        self.prev_hash = prev_hash
        self.prev_index = prev_index
        self.script_sig = Script.parse(script_sig)
        self.sequence = sequence

    def __repr__(self):
        return '{}:{}'.format(
            hexlify(self.prev_hash).decode('ascii'),
            self.prev_index,
        )

    @classmethod
    def parse(cls, stream):
        # Read prev_hash 32 bytes little endian
        prev_hash = stream.read(32)[::-1]

        # Read the prev_index 4 bytes, little endian as int
        prev_index = little_endian_to_int(stream.read(4))

        # Script sig is variable length, use varint to find out its length
        # The first byte signifies the length
        script_sig_length = read_varint(stream)
        script_sig = stream.read(script_sig_length)

        # Read sequence 4 bytes, little endian as int
        sequence = little_endian_to_int(stream.read(4))

        return cls(prev_hash, prev_index, script_sig, sequence)

    def serialize(self):
        result = self.prev_hash[::-1]

        result += int_to_little_endian(self.prev_index, 4)

        script_sig = self.script_sig.serialize()

        result += encode_varint(len(script_sig))

        result += script_sig

        result += int_to_little_endian(self.sequence, 4)

        return result

    @classmethod
    def get_url(self, testnet=True):
        if testnet:
            return 'https://testnet.blockexplorer.com/api'
        else:
            return 'https://btc-bitcore3.trezor.io/api'

    # Fetches Raw TX of the input and creates a Tx Object
    def fetch_tx(self, testnet=True):
        if self.prev_hash not in self.cache:
            url = self.get_url(testnet) + '/rawtx/{}'.format(hexlify(self.prev_hash).decode('ascii'))
            response = requests.get(url)

            if response.status_code == 404:
                raise RuntimeError("404 error received, usually a bad prex_tx hash")

            js_response = response.json()

            if 'rawtx' not in js_response:
                raise RuntimeError('Did not receive expected response: {}'.format(js_response))

            raw = unhexlify(js_response['rawtx'])
            tx = Tx.parse(raw)
            self.cache[self.prev_hash] = tx

        return self.cache[self.prev_hash]

    def value(self, testnet=True):
        tx = self.fetch_tx()

        return tx.tx_outs[self.prev_index].amount

    def der_signature(self, index=0):

        signature = self.script_sig.der_signature(index=index)
        print("Signature in der_signature: {}".format(hexlify(signature)))

        # last byte is the hash_type, rest is the signature
        return signature[:-1], signature[-1]

    def sec_pubkey(self, index=0):
        return self.script_sig.sec_pubkey(index=index)

    # Fetches the script_pub key of this tx input (LOCKING SCRIPT)
    def script_pubkey(self, testnet=True):
        '''Get the scriptPubKey by looking up the tx hash on block explorer server
        Returns the binary scriptpubkey (LOCKING SCRIPT)
        '''
        tx = self.fetch_tx(testnet=testnet)

        return tx.tx_outs[self.prev_index].script_pub_key

    def redeem_script(self):
        ''' Returns a redeem script if there is one '''
        return self.script_sig.redeem_script()



class TxOut:
    def __init__(self, amount, script_pub_key):
        self.amount = amount
        self.script_pub_key = Script.parse(script_pub_key)

    def __repr__(self):
        return '{}:{}'.format(self.amount, self.script_pub_key)

    @classmethod
    def parse(cls, stream):
        # Read amount of the output, Little Endian, 8 bytes as int?
        amount = little_endian_to_int(stream.read(8))

        # Retrieve script_pub_key, variable, read varint for length
        script_pub_key_length = read_varint(stream)
        script_pub_key = stream.read(script_pub_key_length)

        return cls(amount, script_pub_key)

    def serialize(self):
        result = int_to_little_endian(self.amount, 8)

        scriptPubKey = self.script_pub_key.serialize()

        result += encode_varint(len(scriptPubKey))

        result += scriptPubKey

        return result


from unittest import TestCase
from io import BytesIO
from binascii import hexlify, unhexlify


class Signature:
    def __init__(self, r, s):
        if type(s) is not int:
            raise RuntimeError("s is not in int format")

        if type(r) is not int:
            raise RuntimeError("r is not in int format")

        self.s = s
        self.r = r

    @classmethod
    def parse(cls, signature_bin):
        s = BytesIO(signature_bin)

        compound = s.read(1)[0]

        if compound != 0x30:
            raise RuntimeError("Bad Signature")

        length = s.read(1)[0]
        if length + 2 != len(signature_bin):
            raise RuntimeError("Bad Signature Length")

        marker = s.read(1)[0]
        if marker != 0x02:
            raise RuntimeError("Bad Signature")

        rlength = s.read(1)[0]
        r = int(hexlify(s.read(rlength)), 16)
        marker = s.read(1)[0]
        if marker != 0x02:
            raise RuntimeError("Bad Signature")

        slength = s.read(1)[0]
        s = int(hexlify(s.read(slength)), 16)

        # if the length of the signature is into equal to the 6 byte markers, plus r and s, then sig is too long
        if len(signature_bin) != 6 + rlength + slength:
            raise RuntimeError("Signature too long")

        return cls(r, s)

    def der(self):
        # Return r, s in Der Format (serializing)
        rbin = self.r.to_bytes(32, byteorder='big')

        # if rbin has a high bit, add a 00
        if rbin[0] > 128:
            rbin = b'\x00' + rbin

        result = bytes([2, len(rbin)]) + rbin
        sbin = self.s.to_bytes(32, byteorder='big')

        # if sbin has a high bit, add a 00
        if sbin[0] > 128:
            sbin = b'\x00' + sbin

        result += bytes([2, len(sbin)]) + sbin
        return bytes([0x30, len(result)]) + result



class UTXO:
    def __init__(self, tx_hash, block_height, tx_index, value, confirmations, confirmed, double_spend):
        self.tx_hash = tx_hash
        self.block_height = block_height
        self.tx_index = tx_index
        self.value = value
        self.confirmations = confirmations
        self.confirmed = confirmed
        self.double_spend = double_spend

    def __repr__(self):
        return "\nUTXO: \ntx_hash: {}\nblock_height:{}\ntx_index:{}\nvalue:{}\nconfirmations:{}\nconfirmed:{}\ndouble_spend:{}\n".format(
            self.tx_hash,
            self.block_height,
            self.tx_index,
            self.value,
            self.confirmations,
            self.confirmed,
            self.double_spend
        )

    @staticmethod
    def schema_type(raw_utxo):
        if type(raw_utxo) is dict:
            if ('tx_hash' in raw_utxo
                and 'block_height' in raw_utxo  
                and 'tx_input_n' in raw_utxo 
                and 'tx_output_n' in raw_utxo
                and 'value' in raw_utxo 
                and 'ref_balance' in raw_utxo 
                and 'spent' in raw_utxo 
                and 'confirmations' in raw_utxo 
                and 'confirmed' in raw_utxo 
                and 'double_spend' in raw_utxo):
                return 'block_cypher'

        if type(raw_utxo) is dict:
            # ({'hash': 'fea5cbf4efc220a5512d394279778f75937c253cac32c43047cadffc9ee4d85c', 'time': '2018-02-04T12:29:01+0000', 'confirmations': 12468, 'is_coinbase': False, 'value': 156000000, 'index': 1, 'address': 'mfke2PVhGePAy1GfZNotr6LeXfQ5nwnZTa', 'type': 'pubkeyhash', 'multisig': None, 'script': 'OP_DUP OP_HASH160 029692862d60b5f84ba706b37939d074b6c58085 OP_EQUALVERIFY OP_CHECKSIG', 'script_hex': '76a914029692862d60b5f84ba706b37939d074b6c5808588ac'})
            if ('hash' in raw_utxo
                and 'time' in raw_utxo
                and 'confirmations' in raw_utxo
                and 'is_coinbase' in raw_utxo
                and 'value' in raw_utxo
                and 'index' in raw_utxo
                and 'address' in raw_utxo
                and 'type' in raw_utxo
                and 'multisig' in raw_utxo
                and 'script' in raw_utxo
                and 'script_hex' in raw_utxo):
                return 'block_trail'

        raise RuntimeError('Unknown Schema Type') 

    @classmethod
    def parse(cls, raw_utxo):
        schema_type = UTXO.schema_type(raw_utxo)

        # utxo_object = UTXO.parse({'tx_hash': '7c95996721bba829589a622d4bed06410ab455a8be932271d53ec9630b586c20', 'block_height': 1283283, 'tx_input_n': -1, 'tx_output_n': 1, 'value': 104900000, 'ref_balance': 211900000, 'spent': False, 'confirmations': 3326, 'confirmed': '2018-02-17T19:10:32Z', 'double_spend': False})
        if schema_type == 'block_cypher':
            return cls(tx_hash=raw_utxo.get('tx_hash'), 
                    block_height=raw_utxo.get('block_height'), 
                    tx_index=raw_utxo.get('tx_output_n'),
                    value=raw_utxo.get('value'), 
                    confirmations=raw_utxo.get('confirmations'),
                    confirmed=raw_utxo.get('confirmed'), 
                    double_spend=raw_utxo.get('double_spend'))

        # uxto_object = UTXO.parse({'hash': 'fea5cbf4efc220a5512d394279778f75937c253cac32c43047cadffc9ee4d85c', 'time': '2018-02-04T12:29:01+0000', 'confirmations': 12468, 'is_coinbase': False, 'value': 156000000, 'index': 1, 'address': 'mfke2PVhGePAy1GfZNotr6LeXfQ5nwnZTa', 'type': 'pubkeyhash', 'multisig': None, 'script': 'OP_DUP OP_HASH160 029692862d60b5f84ba706b37939d074b6c58085 OP_EQUALVERIFY OP_CHECKSIG', 'script_hex': '76a914029692862d60b5f84ba706b37939d074b6c5808588ac'})
        if schema_type == 'block_trail':
            return cls(tx_hash=raw_utxo.get('hash'),
                    block_height=None,
                    tx_index=raw_utxo.get('index'),
                    value=raw_utxo.get('value'),
                    confirmations=raw_utxo.get('confirmations'),
                    confirmed=raw_utxo.get('time'),
                    double_spend=None)        